// files.cpp - support functions to make filehandling easier
// (c) 1998 Graham Mitchell <mitchell@onr.com> - 14 Feb 98

/*
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (LICENSE.TXT) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation,
	Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if !defined ( __DIR_H )
#include <dir.h>
#endif

#if !defined ( __IOSTREAM_H )
#include <iostream.h>
#endif

#if !defined ( __STRING_H )
#include <string.h>
#endif

#if !defined ( __DOS_H )
#include <dos.h>
#endif

#if !defined ( __CTYPE_H )
#include <ctype.h>
#endif

int fnexpand(char*, const char*);
int copyfile(char*, char*);
void printflags(const int);
inline int exist(const char* filename)	// requires DIR.H
{
	struct ffblk fb;
	return ( findfirst(filename,&fb,0) == 0 );
}

// fnexpand()
//
// Given a filename, generates a fully-qualified pathname for it
// New name is in all-caps, and any .. are removed (if possible)
// Returns an int containing bit flags (as defined in DIR.H) indicating
// what is present in the final expanded filename:
//
// Flag       | Meaning
// -----------+---------------------------------------------------------
// DIRECTORY  | Path includes a directory (and possibly subdirectories)
// DRIVE      | Path includes a drive specification (see dir.h)
// EXTENSION  | Path includes an extension
// FILENAME   | Path includes a file name
// WILDCARDS  | Path contains wildcards (* or ?)
//
// All expanded filenames will contain the bits for DIRECTORY and DRIVE
//
// requires DIR.H, STRING.H, CTYPE.H for islower(), and if compiled with
//	FILES_DEBUG defined, IOSTREAM.H (but not otherwise)
//
int fnexpand(char* result, const char* fragment)
{
	const int CAPS = 0xDF;	// used to mask off "caps" bit from ASCII letter

	int flags, specified_drive;

	char drive[MAXDRIVE];		// MAX* are defined in DIR.H
	char dir[MAXDIR];
	char file[MAXFILE];
	char ext[MAXEXT];
	char* ptr = NULL; 			// for temporary stuff

	// split up fragment and see what we have
	flags = fnsplit(fragment, drive, dir, file, ext);

#ifdef FILES_DEBUG
		cerr << "was passed " << fragment << '\n';
		cerr << "Drive: " << drive << '\n';
		cerr << "Directory: " << dir << '\n';
		cerr << "Filename: " << file << '\n';
		cerr << "Extension: " << ext << '\n';
		cerr << "Wildcards: " << ( (flags&WILDCARDS) ? "Yes\n" : "No\n" );
#endif

	if ( ! (flags & DRIVE) )
	{
		// we need a drive name

#ifdef FILES_DEBUG
			cerr << "No drive. ";
#endif

		drive[0] = getdisk() + 'A';
		drive[1] = ':';
		drive[2] = 0;

#ifdef FILES_DEBUG
			cerr << drive << endl;
#endif
	}

	// get the drive number in case we need it for calls to getcurdir()
	//    we add 1 because getdisk and getcurdir use different drive #s
	//    and capitalize just in case the user didn't
	specified_drive = ( drive[0] & CAPS ) - 'A' + 1;

	if ( !(flags & DIRECTORY) )
	{
		// we need a directory

#ifdef FILES_DEBUG
			cerr << "No directory.  ";
#endif

		// we prefix it with a backslash since getcurdir doesn't

		dir[0] = '\\';
		getcurdir(specified_drive,&dir[1]);

#ifdef FILES_DEBUG
			cerr << dir << endl;
#endif
	}
	else if ( dir[0] != '\\' )
	{
		// we have a directory; but it's specified relative to the cwd

		char prefix[MAXDIR];

#ifdef FILES_DEBUG
			cerr << "Relative directory.  ";
#endif

		// we prefix it with a backslash since getcurdir doesn't

		prefix[0] = '\\';
		getcurdir(specified_drive,&prefix[1]);

		// now append a backslash before we add the relative path
		//	(unless we're currently in root, where we don't need to)

		if ( strcmp(prefix, "\\") != 0 )
			strcat(prefix,"\\");

		// and tack on the relative path to the end

		// leave off the redundant .\ if present
		if ( strncmp(dir, ".\\", 2) == 0 )
			strcat(prefix,dir+2);

		// there's no reason to copy dir if it's just . (cwd)
		else if ( strcmp(dir, ".") != 0 )
			strcat(prefix,dir);

		strcpy(dir, prefix);

#ifdef FILES_DEBUG
			cerr << dir << endl;
#endif
	}

	// Here the directory is fully-specified, but it may contain some
	//  internal previous-directory info (i.e. C:\TEXT\..\SIGS\STUF.TXT),
	//  We'll remove the directory specified to the left of the ..

	ptr = strstr(dir, "..");

	while ( ptr && (ptr-1 != dir) )
	{
#ifdef FILES_DEBUG
		cerr << "Removing '..' from: " << dir;
#endif
		char* trailer = ptr+2;	// points to the backslash after the ..
		ptr-= 2;	// ptr now points just before the preceding backslash

		while ( *ptr != '\\' )
			ptr--;

		// now ptr points at the backslash preceding unneeded directory

		*ptr = 0;		// null-terminate dir[] right here

		strcat(dir, trailer);	// re-append the remainder
#ifdef FILES_DEBUG
		cerr << " -> " << dir << endl;
#endif
		ptr = strstr(dir, "..");	// and check for more
	}

	// Now we put things back together in result[]
	fnmerge(result, drive, dir, file, ext);

	// capitalize everything just to be nice

	int i = 0;
	while ( result[i] != 0 )
	{
		if ( islower(result[i]) )
			result[i] = result[i] & CAPS;
		i++;
	}

	// At this point, all that's left is the case where the user
	// left off the trailing backslash on a directory.

	struct ffblk f;

	// Look up what we have...

	int empty = findfirst(result,&f,16);
	if ( !empty && (f.ff_attrib == 16) && !(flags&WILDCARDS) )

	// it's a directory; append a trailing backslash

		strcat(result,"\\");

	// and we're done; use fnsplit to get our flags

	flags = fnsplit(result, NULL, NULL, NULL, NULL);

	return flags;

}


// copyfile()
//
// Copies file from source to dest; if dest exists, it will be overwritten
//
// requires DOS.H, IOSTREAM.H
//
int copyfile(char* source, char* dest)
{
	const int BUFSIZE = 1024;
	const int O_RDONLY = 1;

	unsigned count, bytes;
	int h_in, h_out;
	char buf[BUFSIZE];

	// create/overwrite destination file
	if (_dos_creat(dest, _A_NORMAL, &h_out) != 0)
	{
		cerr << "\nUnable to create " << dest << endl;
		return -1;
	}
	// open source file for reading
	if (_dos_open(source, O_RDONLY, &h_in) != 0)
	{
		cerr << "\nUnable to open " << source << endl;
		return -1;
	}

	// copy the files 1K at a time
	do {
		_dos_read(h_in, buf, BUFSIZE, &bytes);
		_dos_write(h_out, buf, bytes, &count);
	} while ( bytes == BUFSIZE );

	// close the files
	_dos_close(h_in);
	_dos_close(h_out);

	return 0;
}


// printflags()
//
// Shows the status of flags returned by fnexpand(), useful for debugging
//
// requires DIR.H (for bit flag definitions) and IOSTREAM.H
//
void printflags(const int flags)
{
	cout << "DRIVE  DIR  FILE  EXT  WILD\n  ";
	cout << ( flags&DRIVE ? '*' : ' ' );
	cout << "     ";
	cout << ( flags&DIRECTORY ? '*' : ' ' );
	cout << "    ";
	cout << ( flags&FILENAME ? '*' : ' ' );
	cout << "     ";
	cout << ( flags&EXTENSION ? '*' : ' ' );
	cout << "    ";
	cout << ( flags&WILDCARDS ? '*' : ' ' );
	cout << '\n';

}

